# flake8: noqa
"""
Some code under query/* module is inspired on code created by:
Nick Barrett pointlessrambler@gmail.com
"""

from toute.utils.payload.base import Payload  # noqa
from toute.utils.payload.filters import Filter  # noqa
from toute.utils.payload.queries import Query  # noqa
from toute.utils.payload.aggregates import Aggregate  # noqa
from toute.utils.payload.suggesters import Suggester  # noqa



import elasticsearch.helpers as eh

from six import iteritems, with_metaclass
from toute.bases.py3 import *  # noqa
from toute.bases.document import BaseDocument
from toute.bases.metaclass import ModelMetaclass
from toute.bases.result import ResultSet
from toute.mapping import Mapping
from toute.utils import validate_client
from toute.utils.payload import Payload, Filter
from toute.exceptions import ClientError


class Document(with_metaclass(ModelMetaclass, BaseDocument)):
    """
    Base Document to be extended in your models definitions

    >>> from elasticsearch import Elasticsearch
    >>> from toute import Document, KeywordField
    >>> class MyDoc(Document):
    ...   _autoid = True
    ...   _index = 'indexname'
    ...   _doctype = 'doctypename'
    ...   name = KeywordField()

    >>> obj = MyDoc(name="Gonzo")
    >>> obj.save(es=Elasticsearch())

    >>> MyDoc.filter(name="Gonzo")

    """

    # If _autoid is set to False the id Field will not be automatically
    # included in the Document model and you will need to specify a field
    # called 'id' preferably a KeywordField
    _autoid = True

    # _validators is a list of callable, each one executed receiving the
    # document instance, and should return None
    # else document is invalid and will not be saved
    # to invalidate the callable should raise validationError or return value
    _validators = None

    @classmethod
    def having(cls, **kwargs):
        meta_attributes = ['index', 'doctype', 'es', 'autoid', 'validators',
                           'strict', 'fields']
        for k, v in kwargs.items():
            setattr(cls, "_" + k if k in meta_attributes else k, v)
        return cls

    @classmethod
    def get_es(cls, es):
        """
        This proxy-method allows the client overwrite
        and the use of a default client for a document.
        Document transport methods should use cls.get_es(es).method()
        This method also validades that the connection is a valid ES client.

        :param es: The Es client or None
        :return: elasticsearch.ElasticSearch() instance or equivalent client
        """
        if not es and hasattr(cls, '_es'):
            es = cls._es if not callable(cls._es) else cls._es()
        validate_client(es)
        return es

    @classmethod
    def refresh(cls, es=None):
        """
        Used to refresh an index and its shards
        Utility for tests purposes

        :param es: ES client
        :return: ES Metadata
        """
        return cls.get_es(es).indices.refresh()

    def save(self, es=None):
        """
        Save current instance of a Document

        >>> obj = Document(field='value')
        >>> obj.save()

        :param es: ES client or None (if implemented a default in Model)
        :return: Es meta data
        """
        doc = self.to_dict()
        saved_document = self.get_es(es).index(
            index=self._index,
            doc_type=self._doctype,
            id=self.id,  # noqa
            body=doc
        )
        created = saved_document.get('created')
        if created:
            self.id = saved_document['_id']
        return saved_document

    def update(self, body=None, es=None, meta=None, **kwargs):
        """
        Update a single document

        Using fields
        >>> Document().update(some_field="some_value")

        Using a body dict
        >>> Document().update({'some_field': "some_value"})

        Or a script
        >>> Document().update(script="for(x in data){x}",
        ...                   lang="groovy",
        ...                   params={'data': [...]})

        :param es: ES client
        :param meta: Extra values to be passed to client
        :param body: Optional values passed as dict
        :param kwargs: values to change
        :return: Update result
        """
        body = body or {}
        body.update(kwargs)
        updated_data = self.update_by_id(
            self.id, body=body, es=es, meta=meta
        )
        if 'script' not in body:
            for key, value in iteritems(body):
                setattr(self, key, value)
        return updated_data

    @classmethod
    def update_by_id(cls, doc_id, body=None, es=None, meta=None, **kwargs):
        """
        Update a single document using its id on BaseClass

        Using fields
        >>> Document.update_by_id(1234, some_field="some_value")
        Using boy dict
        >>> Document.update_by_id(1234, {'some_field': 'some_value'})

        Or a script
        >>> Document.update_by_id(1234,
        ...                       script="for(x in data){x}",
        ...                       lang="groovy",
        ...                       params={'data': [...]})

        :param doc_id: The document of the id to be updated
        :param body: Optional values passed as dict
        :param es: ES client
        :param meta: Extra values to be passed to client
        :param kwargs: values to change
        :return: Update result
        """
        body = body or {}
        body.update(kwargs)
        meta = meta or {}

        if 'script' not in body and 'doc' not in body:
            body = {'doc': body}

        updated_data = cls.get_es(es).update(
            index=cls._index,
            doc_type=cls._doctype,
            id=doc_id,  # noqa
            body=body,
            **meta
        )
        return updated_data

    def delete(self, es=None):
        """
        Delete current instance of a Document

        >>> obj = Document.get(id=123)
        >>> obj.delete()

        :param es: ES client or None (if implemented a default in Model)
        :return: ES meta data
        """
        return self.get_es(es).delete(
            index=self._index,
            doc_type=self._doctype,
            id=self.id,  # noqa
        )

    @classmethod
    def create(cls, es=None, **kwargs):
        """
        Creates and returns an instance of the Document

        >>> Document.create(field='value')
        <Document: {'field': 'value'}>

        :param es: ES client or None (if implemented a default in Model)
        :param kwargs: fields and its values
        :return: Instance of the Document created
        """
        instance = cls(**kwargs)
        instance.save(es)
        return instance

    @classmethod
    def all(cls, *args, **kwargs):
        """
        Returns a ResultSet with all documents without filtering
        A semantic shortcut to filter() without keys

        :param: <See filter parameters>
        :return: A ResultSet with all documents in the index/type
        """
        return cls.filter(*args, **kwargs)

    @classmethod
    def exists(cls, id, es=None, **kwargs):  # noqa
        """
        Tell if document exists on index

        >>> Document.exists(id=123)

        :param id: The _id or _uid of the object
        :param es: ES client or None (if implemented a default in Model)
        :param kwargs: extra key=value to be passed to es client
        :return: True or False
        """
        return cls.get_es(es).exists(
            index=cls._index,
            doc_type=cls._doctype,
            id=id,
            **kwargs
        )

    @classmethod
    def get(cls, id, es=None, **kwargs):  # noqa
        """
        A get query returning a single document by _id or _uid

        >>> Document.get(id=123)

        :param id: The _id or _uid of the object
        :param es: ES client or None (if implemented a default in Model)
        :param kwargs: extra key=value to be passed to es client
        :return: A single Doc object
        """
        es = cls.get_es(es)
        res = es.get(index=cls._index,
                     doc_type=cls._doctype,
                     id=id,
                     **kwargs)
        return cls.from_es(res)

    @classmethod
    def count_by_query(cls, *args, **kwargs):
        """
        Count documents using a specific raw query
        example: Counting all documents having non-null name field

        >>> query = {
        ...     "query": {
        ...         "filtered": {
        ...             "query": {"match_all": {}},
        ...             "filter": {"exists": {"field": "name"}}
        ...         }
        ...     }
        ... }
        >>> total = Document.count_by_query(query)

        :param args: <see .count parameters>
        :param kwargs: <see .count parameters>
        :return: Integer count
        """
        return cls.count(_method='search', *args, **kwargs)

    @classmethod
    def count(cls, _method='filter', *args, **kwargs):
        """
        Count documents by query or all if no param
        :param args: <see .filter parameters>
        :param _method: filter or search
        :param kwargs: <see .filter parameters>
        :return: Integer count
        """
        kwargs['perform_count'] = True
        return getattr(cls, _method)(*args, **kwargs)

    @classmethod
    def filter(cls, es=None, ids=None,
               size=None, perform_count=False, **filters):
        """
        A match_all query with filters

        >>> Document.filter(ids=[123, 456])
        >>> Document.filter(name="Gonzo", city="Tunguska", size=10)

        :param es: ES client or None (if implemented a default in Model)
        :param ids: Filtering by _id or _uid
        :param size: size of result, default 100
        :param filters: key=value parameters
        :param perform_count: If True, dont return objects, only count
        :return: Iterator of Doc objets
        """

        es = cls.get_es(es)

        if ids and filters:
            raise ValueError(
                "You can't specify ids together with other filters"
            )

        if ids:
            query = {
                "query": {
                    "filtered": {
                        "query": {"match_all": {}},
                        "filter": {"ids": {"values": list(ids)}}
                    }
                }
            }
        elif filters:
            query = {
                "query": {
                    "bool": {
                        "must": [
                            {"match": {key: value}}
                            for key, value in filters.items()
                        ]
                    }
                }
            }
        else:
            query = {
                "query": {
                    "match_all": {}
                }
            }

        size = len(ids) if ids else size
        search_args = dict(
            index=cls._index,
            doc_type=cls._doctype,
            body=query
        )

        if perform_count:
            return es.count(**search_args)['count']

        if size:
            search_args['size'] = size

        resp = es.search(**search_args)
        return cls.build_result(resp, es=es, query=query, size=size)

    @classmethod
    def search(cls, query, es=None, perform_count=False, **kwargs):
        """
        Takes a raw ES query in form of a dict or Payload and
        return Doc instances iterator
        >>> query = {
        ...     "query": {
        ...        "bool": {
        ...            "must": [
        ...                {"match": {"name": "Gonzo"}}
        ...            ]
        ...        }
        ...    }
        ...}
        >>> results = Document.search(query, size=10)

        :param query: raw_query(preferable) or Query or Payload instance
        :param es: ES client or None (if implemented a default in Model)
        :param perform_count: If True, dont return objects, only count
        :param kwargs: extra key=value to be passed to es client
        :return: Iterator of Doc objets

        NOTE: Checking istance types is expensive, please prefer to use
        raw queries ex:
        Document.search({"query": ...}) || .search(payload_instance.dict)
        """

        if not isinstance(query, dict):
            # if not a raw dict query
            if isinstance(query, Payload):  # must be a Payload instance
                query = query.dict
            elif isinstance(query, Filter):  # must be a Filter
                query = Payload(filter=query).dict
            else:  # or a Query to wrap
                query = Payload(query=query).dict

        es = cls.get_es(es)
        search_args = dict(
            index=cls._index,
            doc_type=cls._doctype,
            body=query,
            **kwargs
        )

        if perform_count:
            return es.count(**search_args)['count']

        return cls.build_result(
            es.search(**search_args),
            es=es,
            query=query,
            size=kwargs.get('size')
        )

    @classmethod
    def build_result(cls, resp, query=None, es=None, size=None):
        """
        Takes ES client response having ['hits']['hits']
        and turns it to an generator of Doc objects
        :param resp: ES client raw results
        :param query: The query used to build the results
        :param es: Es client
        :param size: size of results
        :return: ResultSet: a generator of Doc objects
        """

        if resp.get('timed_out'):
            raise ClientError("Timeout")

        return ResultSet(
            resp=resp,
            model=cls,
            query=query,
            size=size,
            es=cls.get_es(es)
        )

    @classmethod
    def save_all(cls, docs, es=None, **kwargs):
        """
        Save various Doc instances in bulk

        >>> docs = (Document(value=value) for value in [1, 2, 3])
        >>> Document.save_all(docs)

        :param docs: Iterator of Document instances
        :param es: ES client or None (if implemented a default in Model)
        :param kwargs: Extra params to be passed to streaming_bulk
        :return: ES metadata
        """
        actions = [
            {
                '_op_type': 'index',
                '_index': cls._index,
                '_type': cls._doctype,
                '_id': doc.id,
                '_source': doc.to_dict()
            }
            for doc in docs
        ]
        return eh.bulk(cls.get_es(es), actions, **kwargs)

    @classmethod
    def update_all(cls, docs, es=None, meta=None, **kwargs):
        """
        Update various Doc instances in bulk

        >>> docs = (Document(value=value) for value in [1, 2, 3])
        # change all values to zero
        >>> Document.update_all(docs, value=0)

        :param docs: Iterator of Document instances
        :param es: ES client or None (if implemented a default in Model)
        :param meta: Extra values to be passed to client
        :param kwargs: Extra params to be passed to streaming_bulk
        :return: Es Metadata
        """
        actions = (
            {
                '_op_type': 'update',
                '_index': cls._index,
                '_type': cls._doctype,
                '_id': doc.id,
                'doc': kwargs
            }
            for doc in docs
        )
        return eh.bulk(cls.get_es(es), actions, **meta if meta else {})

    @classmethod
    def delete_all(cls, docs, es=None, **kwargs):
        """
        Delete various Doc instances in bulk

        >>> docs = (Document(value=value) for value in [1, 2, 3])
        >>> Document.delete_all(docs)

        :param docs: Iterator of Document instances or a list of ids
        :param es: ES client or None (if implemented a default in Model)
        :param kwargs: Extra params to be passed to streaming_bulk
        :return: ES metadata
        """
        actions = [
            {
                '_op_type': 'delete',
                '_index': cls._index,
                '_type': cls._doctype,
                '_id': getattr(doc, 'id', doc),
            }
            for doc in docs
        ]
        return eh.bulk(cls.get_es(es), actions, **kwargs)

    @classmethod
    def random(cls, size=None):
        _query = {
            "query": {
                "function_score": {
                    "query": {"match_all": {}},
                    "random_score": {}
                }
            }
        }
        results = cls.search(_query, size=size)
        return results

    @classmethod
    def put_mapping(cls, *args, **kwargs):
        """
        If index does not exist it is created with mapping
        If exists mapping is updated
        :return: acknowlege
        """
        mapping = Mapping(cls, *args, **kwargs)
        return mapping.save()

    @classmethod
    def init(cls, *args, **kwargs):
        return {
            'mapping': cls.put_mapping(*args, **kwargs),
            'settings': 'Not Implemented yet',
            'analysers': 'Not Implemented yet'
        }

    def __unicode__(self):
        return unicode(self.__str__())

    def __str__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self.to_dict())



# -*- coding: utf-8 -*-
import logging
import math
from copy import deepcopy
from six.moves import range
from toute.exceptions import PaginationError, StopPagination


class Pagination(object):

    def __init__(self, iterable, page=1, per_page=10):
        """
        Initialize an iterator
        :param iterable: Payload (recommended), ResultSet or an iterator
        :param page:
        :param per_page:
        :return:
        """
        self.init(iterable, page, per_page)

    def init(self, iterable, page, per_page):
        page = int(page or 1)
        per_page = int(per_page or 10)
        logging.getLogger(__name__).info(page)
        # if page < 1:
        #     raise PaginationError("Page is lower than 1")

        self.iterable = iterable  # noqa
        self.page = page  # noqa
        self.per_page = per_page  # noqa

        if hasattr(iterable, 'count'):
            self.total = self.total_size = int(iterable.count())
        else:
            self.total = self.total_size = len(iterable)  # noqa

        start_index = (page - 1) * per_page
        end_index = page * per_page

        if hasattr(iterable, 'search'):  # it is a Payload
            struct_bck = deepcopy(iterable._struct)
            # apply pagination
            total_size = iterable._struct.get('size')
            if total_size:
                self.total_size = int(total_size)  # noqa
            iterable.from_(start_index)
            iterable.size(per_page)
            self.items = iterable.search()
            # restore Payload state
            iterable._struct = struct_bck
        else:
            self.items = iterable[start_index:end_index]  # noqa

        if not self.items and page != 1:
            raise StopPagination("There is no items to paginate")

        if self.page > self.pages:
            raise StopPagination("Pagination Overflow")

    def count(self):
        """
        The minimum between search.count and specified total_size
        :return: integer
        """
        return min(self.total, self.total_size)

    @property
    def pages(self):
        """The total number of pages"""
        return int(math.ceil(self.count() / float(self.per_page)))

    def prev_page(self, inplace=False):
        """Returns a :class:`Pagination` object for the previous page."""
        if self.iterable is None:
            raise PaginationError('iterable is needed')
        if not self.has_prev:
            raise StopPagination("There is no previous page")
        return (
            self.__class__
            if not inplace else
            self.init
        )(self.iterable, self.page - 1, self.per_page)

    def backward(self):
        return self.prev_page(inplace=True)

    @property
    def prev_num(self):
        """Number of the previous page."""
        if self.has_prev:
            return self.page - 1

    @property
    def has_prev(self):
        """True if a previous page exists"""
        return self.page > 1

    def next_page(self, inplace=False):
        """Returns a :class:`Pagination` object for the next page."""
        if self.iterable is None:
            raise PaginationError('iterable is needed')
        if not self.has_next:
            raise StopPagination("There is no next page")
        return (
            self.__class__
            if not inplace else
            self.init
        )(self.iterable, self.page + 1, self.per_page)

    def forward(self):
        self.next_page(inplace=True)

    @property
    def has_next(self):
        """True if a next page exists."""
        return self.page < self.pages

    @property
    def next_num(self):
        """Number of the next page"""
        if self.has_next:
            return self.page + 1

    def iter_pages(self, left_edge=2, left_current=2,
                   right_current=5, right_edge=2):
        """Iterates over the page numbers in the pagination.  The four
        parameters control the thresholds how many numbers should be produced
        from the sides.  Skipped page numbers are represented as `None`.
        This is how you could render such a pagination in the templates:

        .. sourcecode:: html+jinja

            {% macro render_pagination(pagination, endpoint) %}
              <div class=pagination>
              {%- for page in pagination.iter_pages() %}
                {% if page %}
                  {% if page != pagination.page %}
                    <a href="{{ url_for(endpoint, page=page) }}">{{ page }}</a>
                  {% else %}
                    <strong>{{ page }}</strong>
                  {% endif %}
                {% else %}
                  <span class=ellipsis>…</span>
                {% endif %}
              {%- endfor %}
              </div>
            {% endmacro %}
        """
        last = 0
        for num in range(1, self.pages + 1):
            if num <= left_edge or \
                    (
                        num > self.page - left_current - 1 and  # noqa
                        num < self.page + right_current  # noqa
                    ) or num > self.pages - right_edge:
                if last + 1 != num:
                    yield None
                yield num
                last = num

    @property
    def meta(self):
        return {
            'total': self.count(),
            'pages': self.pages,
            'per_page': self.per_page,
            'page': self.page,
            'next_page': self.next_num,
            'previous_page': self.prev_num
        }

    def to_dict(self):
        return {
            "items": self.items.to_dict(),
            "meta": self.meta
        }



# coding: utf-8
import time
import elasticsearch.helpers as eh
from six import text_type

HITS = 'hits'


class ResultSet(object):
    def __init__(self, resp, model, query=None,
                 size=None, es=None, meta=None):
        self._model = model
        self._values = self._hits = resp.get(HITS, {}).get(HITS, [])
        self._query = query
        self._es = model.get_es(es)
        self._size = size or len(self._values)
        self._meta = self._extract_meta(resp)
        if meta:
            self._meta.update(meta)
        self._all_values = []

    def __iter__(self):
        return self.values

    def _extract_meta(self, resp):
        meta = {key: resp[key] for key in resp if key != HITS}
        if HITS in resp:
            hits = resp[HITS]
            meta[HITS] = {key: hits[key] for key in hits if key != HITS}
        return meta

    @property
    def meta(self):
        return self._meta

    @property
    def values(self):
        return (
            self._model.from_es(hit=hit)
            for hit in self._hits
        )

    @property
    def all_values(self):
        if not self._all_values:
            self._all_values = [i for i in self.values]
        return self._all_values

    def __getitem__(self, item):
        return self.all_values[item]

    def reload(self, sleep=1):
        time.sleep(sleep)
        self._all_values = []
        resp = self._es.search(
            index=self._model._index,
            doc_type=self._model._doctype,
            body=self._query,
            size=self._size or len(self._values)
        )
        self._hits = self._values = resp.get('hits', {}).pop('hits', [])
        self._meta = resp
        return resp

    def update(self, meta=None, **kwargs):
        if kwargs:
            actions = [
                {
                    '_op_type': 'update',
                    '_index': self._model._index,
                    '_type': self._model._doctype,
                    '_id': doc.id,
                    'doc': kwargs
                }
                for doc in self.values
            ]
            return eh.bulk(self._es, actions, **meta if meta else {})

    def delete(self, meta=None, **kwargs):
        actions = (
            {
                '_op_type': 'delete',
                '_index': self._model._index,
                '_type': self._model._doctype,
                '_id': doc.id,
            }
            for doc in self.values
        )
        return eh.bulk(self._es, actions, **meta if meta else {})

    def count(self):
        return min(self._size, self.meta.get('hits', {}).get('total')['value'])

    def to_dict(self, *args, **kwargs):
        """
        returns a list of Documents transformed in dicts
        [{}, {}, ...]
        :param args: passed to item
        :param kwargs: passed to item
        :return:
        """
        return [item.to_dict(*args, **kwargs) for item in self.values]

    def get_values(self, *fields):
        """
        if args is only one field .get_values('id') return a list of lists
        [123, 456, 789]
        If args is more than one field return a list of tuples
        .get_values("id", "name")
        [(123, "John"), (789, "mary"), ...]
        :param fields: a list of fields
        :return:
        """
        if not fields:
            raise AttributeError("At least one field is required")

        if len(fields) > 1:
            return [
                tuple(getattr(value, field) for field in fields)
                for value in self.values
            ]
        else:
            return [getattr(value, fields[0]) for value in self.values]

    def __unicode__(self):
        return text_type(self.__unicode__())

    def __str__(self):
        return "<ResultSet: {i.values}>".format(i=self)



from toute.utils.payload.meta_util import (
    make_struct, unroll_definitions, unroll_struct
)


class MetaFilterQuery(type):
    def __init__(self, name, bases, d):
        super(MetaFilterQuery, self).__init__(name, bases, d)
        unroll_definitions(self._definitions)

    def __getattr__(self, key):
        if key == '__test__':
            return None

        self._validate_key(key)
        return lambda *args, **kwargs: self(
            key,
            make_struct(self._definitions[key], *args, **kwargs)
        )

    def _validate_key(self, key):
        if key != "__slots__" and key not in self._definitions:
            raise self._exception(key)


class MetaAggregate(MetaFilterQuery):
    def __getattr__(self, key):
        if key == '__test__':
            return None

        self._validate_key(key)
        return lambda *args, **kwargs: self(
            key,
            args[0],
            make_struct(self._definitions[key], *args[1:], **kwargs)
        )


class MetaSuggester(MetaFilterQuery):
    def __getattr__(self, key):
        if key == '__test__':
            return None

        self._validate_key(key)
        return lambda *args, **kwargs: self(
            key,
            args[0],
            args[1],
            make_struct(self._definitions[key], *args[2:], **kwargs)
        )


class BaseFilterQuery(object):
    _struct = None
    _dsl_type = None

    def __init__(self, dsl_type, struct):
        self._dsl_type = dsl_type
        self._struct = struct

    @property
    def dict(self):
        return self.as_dict()

    def as_dict(self):
        # Handle reserved Python keyword alternatives (from_, or_)
        dsl_type = (
            self._dsl_type[:-1]
            if self._dsl_type.endswith('_')
            else self._dsl_type
        )
        return {dsl_type: unroll_struct(self._struct)}


class BaseAggregate(BaseFilterQuery):
    _name = None

    def __init__(self, dsl_type, name, struct):
        self._dsl_type = dsl_type
        self._struct = struct
        self._name = name
        self._aggs = []

    def as_dict(self):
        struct = {
            self._name: {
                self._dsl_type: unroll_struct(self._struct)
            }
        }

        if self._aggs:
            aggregates = {}

            for agg in self._aggs:
                aggregates.update(agg.as_dict())

            struct[self._name]['aggregations'] = aggregates

        return struct

    def aggregate(self, *aggregates):
        self._aggs.extend(aggregates)
        return self


class BaseSuggester(BaseFilterQuery):
    _name = None

    def __init__(self, dsl_type, name, text, struct):
        self._dsl_type = dsl_type
        self._struct = struct
        self._name = name
        self._text = text
        self._suggs = []

    def as_dict(self):
        struct = {
            self._name: {
                "text": self._text,
                self._dsl_type: unroll_struct(self._struct)
            }
        }

        if self._suggs:
            for sugg in self._suggs:
                struct.update(sugg.as_dict())

        return struct



from toute.fields import KeywordField
from toute.bases.field import BaseField

from six import iteritems


class ModelMetaclass(type):

    def __new__(mcls, name, bases, attrs):  # noqa
        attrs['_fields'] = {}
        for base in bases:
            if hasattr(base, '_autoid'):
                if base._autoid and 'id' not in attrs:
                    attrs['id'] = KeywordField(field_name='id')
                break

        for base in bases:
            for key, value in iteritems(base.__dict__):
                if isinstance(value, BaseField):
                    value._field_name = key
                    attrs['_fields'][key] = value

        for key, value in iteritems(attrs):
            if isinstance(value, BaseField):
                value._field_name = key
                attrs['_fields'][key] = value
        cls = type.__new__(mcls, name, bases, attrs)
        if any(x.__name__ == 'EmbeddedDocument' for x in bases):
            cls._type = cls
        return cls



import collections
import logging


class Mapping(object):
    """
    Used to generate mapping based in document field definitions

    >>> class Obj(Document):
    ...     name = KeywordField()

    And you can use a Mapping to refresh mappings
    (use in cron jobs or call periodically)

    obj_mapping = Mapping(Obj)
    obj_mapping.save()

    Adicionally this class handle index settings configuration. However this
    operation must be done at elasticsearch index creation.

    """
    def __init__(self, document_class=None, enable_all=True):
        self.document_class = document_class
        self.enable_all = enable_all

    def _generate(self, doc_class):
        """
        Generate the mapping acording to doc_class.

        Args:
            doc_class: toute.Document object containing the model to be
            mapped to elasticsearch.
        """
        m = {
            doc_class._doctype: {
                "properties": {
                    field_name: field_instance.mapping
                    for field_name, field_instance in doc_class._fields.items()
                    if field_name != "id"
                }
            }
        }
        logging.getLogger(__name__).info(m)
        return m

    def generate(self):
        return self._generate(self.document_class)

    def save(self, es=None):
        """
        Save the mapping to index.

        Args:
            es: elasticsearch client intance.
        """
        es = self.document_class.get_es(es)
        if not es.indices.exists(index=self.document_class._index):
            return es.indices.create(
                index=self.document_class._index,
                body={"mappings": self.generate()},
                params={"include_type_name": "true"}
            )

    def build_configuration(self, models_to_mapping, custom_settings, es=None):
        """
        Build request body to add custom settings (filters, analizers, etc) to index.

        Build request body to add custom settings, like filters and analizers,
        to index.

        Args:
            models_to_mapping: A list with the toute.Document objects that
            we want generate mapping.

            custom_settings: a dict containing the configuration that will be
            sent to elasticsearch/_settings (www.elastic.co/guide/en/
                elasticsearch/reference/current/indices-update-settings.html)

            es: elasticsearch client intance.
        """  # noqa
        indexes = set()
        configuration = {}
        mapped_models = [x for x in models_to_mapping]
        for model in mapped_models:
            indexes.add(model._index)
            es = model.get_es(es)
        for index in indexes:
            if es.indices.exists(index=index):
                msg = 'Settings are supported only on index creation'
                raise ValueError(msg)
        mappings_by_index = collections.defaultdict(dict)
        for model in mapped_models:
            mapping = self._generate(model)
            mappings_by_index[model._index].update(mapping)
        for index, mappings in mappings_by_index.items():
            settings = {
                "settings": custom_settings,
                "mappings": mappings
            }
            configuration[index] = settings
        return configuration

    def configure(self, models_to_mapping, custom_settings=None, es=None):
        """
        Add custom settings like filters and analizers to index.

        Add custom settings, like filters and analizers, to index. Be aware
        that elasticsearch only allow this operation on index creation.

        Args:
            models_to_mapping: A list with the toute.Document objects that
            we want generate mapping.

            custom_settings: a dict containing the configuration that will be
            sent to elasticsearch/_settings (www.elastic.co/guide/en/
                elasticsearch/reference/current/indices-update-settings.html)

            es: elasticsearch client intance.
        """
        if not isinstance(models_to_mapping, collections.Iterable):
            raise AttributeError('models_to_mapping must be iterable')

        if custom_settings:
            for model in models_to_mapping:
                es = model.get_es(es)
                if es:
                    break
            configurations = self.build_configuration(
                models_to_mapping,
                custom_settings,
                es
            )
            for index, settings in configurations.items():
                es.indices.create(index=index, body=settings)
        else:
            mapped_models = [x for x in models_to_mapping]
            for model in mapped_models:
                model.put_mapping()



from collections import Iterable

from toute.bases.field import BaseField
from toute.bases.metaclass import ModelMetaclass
from toute.exceptions import RequiredField, InvalidMultiField
from toute.exceptions import FieldTypeMismatch

from six import with_metaclass, iteritems


class EmbeddedDocument(with_metaclass(ModelMetaclass, BaseField)):

    def _to_dict_element(self, real_obj):
        result = {}
        for field_name, field_class in iteritems(self._fields):
            value = getattr(real_obj, field_name)
            result.update({field_name: field_class.to_dict(value)})
        return result

    def to_dict(self, value):
        if value is not None:
            if self._multi:
                return [self._to_dict_element(elem) for elem in value]
            return self._to_dict_element(value)

    def _validate_element(self, elem):
        if not isinstance(elem, EmbeddedDocument):
            raise FieldTypeMismatch(self._field_name, self.__class__._type,
                                    elem.__class__)
        for field_name, field_class in iteritems(self._fields):
            value = getattr(elem, field_name)
            field_class.validate(value)

    def validate(self, value):
        if value is None:
            if self._required:
                raise RequiredField(self._field_name)
        else:
            if self._multi:
                if not isinstance(value, Iterable):
                    raise InvalidMultiField(self._field_name)
                for elem in value:
                    self._validate_element(elem)
            else:
                self._validate_element(value)

    def _from_dict_element(self, dct):
        params = {}
        for field_name, field_class in iteritems(self._fields):
            serialized = dct.get(field_name)
            value = field_class.from_dict(serialized)
            params[field_name] = value
        return self.__class__(**params)

    def from_dict(self, serialized):
        if serialized is None:
            return None
        if self._multi:
            return [self._from_dict_element(elem) for elem in serialized]
        return self._from_dict_element(serialized)



from toute.utils.payload.meta import BaseFilterQuery, MetaFilterQuery
from toute.utils.payload.exception import NoFilter

from six import with_metaclass


FILTERS = {
    'and_': ['_filter'],
    'bool': {
        'kwargs': ({('must', 'must_not', 'should'): ['_filter']},)
    },
    'exists': {
        'args': ('field',)
    },
    'geo_bounding_box': {
        'field': True,
        'kwargs': ('top_left', 'bottom_right')
    },
    'geo_distance': {
        'field': True,
        'kwargs': ('lat', 'lon')
    },
    'geo_distance_range': {
        'field': True,
        'kwargs': ('lat', 'lon')
    },
    'geo_polygon': {
        'field': True,
        'args': ({'points': []},)
    },
    'geo_shape': {
        'field': True,
        'kwargs': ('type', {'coordinates': []}),
        'field_process': lambda q: {'shape': q}
    },
    'geohash_shell': {
        'field': True,
        'kwargs': ('lat', 'lon',)
    },
    'has_child': {
        'args': ('type',),
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'has_parent': {
        'args': ('parent_type',),
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'ids': {
        'args': ({'values': []},),
        'kwargs': ('type',)
    },
    'indices': {
        'args': ({'indices': []},),
        'kwargs': ({('filter', 'no_match_filter'): '_filter'},)
    },
    'limit': {
        'args': ('value',)
    },
    'match_all': {},
    'missing': {
        'args': ('field',)
    },
    'nested': {
        'args': ('path', {'filter': '_filter'}),
    },
    'not_': {
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'or_': ['_filter'],
    'prefix': {
        'field': True,
        'args': ('value',)
    },
    'range': {
        'field': True,
        'kwargs': ('gte', 'gt', 'lte', 'lt')
    },
    'regexp': {
        'field': True,
        'args': ('value',),
        'kwargs': ('flags', 'max_determinized_states')
    },
    'script': {
        'args': ('script',)
    },
    'term': {
        'field': True,
        'args': ('value',)
    },
    'terms': {
        'field': True,
        'value_only': True,
        'args': ({'value': []},)
    },
    'type': {
        'args': ('value',)
    }
}


class Filter(with_metaclass(MetaFilterQuery, BaseFilterQuery)):
    _ee_type = 'filter'
    _definitions = FILTERS
    _exception = NoFilter

    @classmethod
    def query(cls, query, cache=False):
        if cache:
            return cls('fquery', {
                'query': query,
                '_cache': True
            })
        else:
            return cls('query', query)



from toute.bases.py3 import *  # noqa
from toute.fields import KeywordField
from toute.exceptions import ValidationError

import warnings
from six import iteritems


class BaseDocument(object):
    _strict = False
    _validators = None
    _query_fields = None

    def _initialize_defaults_fields(self, ignore=None):
        ignore = ignore or []
        for key, field_instance in iteritems(self.__class__._fields):
            if key not in ignore:
                default = self.get_default_value_for_field(field_instance)
                setattr(self, key, default)

    def get_default_value_for_field(self, field_instance):
        default = field_instance._default
        if callable(default):
            try:
                default = field_instance._default(self, field_instance)
            except TypeError:
                default = field_instance._default()
        return default

    def __init__(self, *args, **kwargs):
        klass = self.__class__.__name__
        if not hasattr(self, '_doctype'):
            raise ValueError('{} have no _doctype attribute'.format(klass))
        if not hasattr(self, '_index'):
            raise ValueError('{} have no _index attribute'.format(klass))
        id_field = self.__class__._fields.get("id")
        if id_field and not isinstance(id_field, KeywordField):
            warnings.warn(
                'To avoid mapping problems, '
                'it is recommended to define the id field as a KeywordField'
            )
        for key, value in iteritems(kwargs):
            setattr(self, key, value)
        self._initialize_defaults_fields(ignore=kwargs.keys())

    def __setattr__(self, key, value):
        if (not key.startswith('_')) and key not in self._fields:
            raise KeyError('`{}` is an invalid field'.format(key))
        field_instance = self._fields.get(key)
        if field_instance and not self._strict:
            value = field_instance.from_dict(value)
        super(BaseDocument, self).__setattr__(key, value)

    def to_dict(self, validate=True, only=None, exclude=None):
        """
        Transform value from Python to Dict to be saved in E.S
        :param validate: If should validate before transform
        :param only: if specified only those fields will be included
        :param exclude: fields to exclude from dict
        :return: dict
        """
        if validate:
            self.validate()

        if only:
            fields = {
                k: v for k, v in iteritems(self._fields)
                if k in only
            }
        elif exclude:
            fields = {
                k: v for k, v in iteritems(self._fields)
                if k not in exclude
            }
        else:
            fields = self._fields

        return {
            field_name: field_instance.to_dict(
                getattr(self, field_name), validate=validate
            )
            for field_name, field_instance in iteritems(fields)
        }

    @classmethod
    def from_dict(cls, dct):
        """
        Transform data read from E.S to Python Document Object
        :param dct: Result from E.S (hits, source as dict)
        :return: Instance of Document
        """
        params = {}
        for field_name, field_instance in iteritems(cls._fields):
            serialized = dct.get(field_name)
            value = field_instance.from_dict(serialized)
            params[field_name] = value
        return cls(**params)

    @classmethod
    def from_es(cls, hit):
        """
        Takes E.S hit element containing
        [u'_score', u'_type', u'_id', u'_source', u'_index']

        :param hit: E.S hit
        :return: Document instance
        """
        instance = cls.from_dict(dct=hit.get('_source', {}))
        instance._id = instance.id = hit.get('_id')
        instance._score = hit.get('_score')
        instance._query_fields = hit.get('fields', None)
        return instance

    def validate(self):
        if self._validators:
            for validator in self._validators:
                """
                Functions in self._validators receives document instance
                should return None or
                raise Exception (ValidationError) or return any value
                """
                val = validator(self)
                if val:
                    raise ValidationError("Invalid: %s" % val)



from toute.utils.payload.meta import BaseAggregate, MetaAggregate
from toute.utils.payload.exception import NoAggregate

AGGREGATES = {
    'min': {
        'args': ('field',)
    },
    'max': {
        'args': ('field',)
    },
    'sum': {
        'args': ('field',)
    },
    'avg': {
        'args': ('field',)
    },
    'stats': {
        'args': ('field',)
    },
    'extended_stats': {
        'args': ('field',)
    },
    'value_count': {
        'args': ('field',)
    },
    'percentiles': {
        'args': ('field',)
    },
    'percentile_ranks': {
        'args': ('field',)
    },
    'cardinality': {
        'args': ('field',)
    },
    'geo_bounds': {
        'args': ('field',)
    },
    'top_hits': {
    },
    'scripted_metric': {
    },
    'global': {
    },
    'filter': {
        'args': ({'filter': '_filter'},)
    },
    'filters': {
        'args': ({'filters': ['_filter']},)
    },
    'missing': {
        'args': ('field',)
    },
    'nested': {
        'args': ('path',)
    },
    'reverse_nested': {
    },
    'children': {
        'args': ('type',)
    },
    'terms': {
        'args': ('field',)
    },
    'significant_terms': {
        'args': ('field',)
    },
    'range': {
        'args': ('field', {'ranges': []})
    },
    'date_range': {
        'args': ('field', {'ranges': []})
    },
    'ip_range': {
        'args': ('field', {'ranges': []})
    },
    'histogram': {
        'args': ('field', 'interval')
    },
    'date_histogram': {
        'args': ('field', 'interval')
    },
    'geo_distance': {
        'args': ('field', 'origin', {'ranges': []})
    },
    'geohash_grid': {
        'args': ('field',)
    }
}


class Aggregate(BaseAggregate):
    __metaclass__ = MetaAggregate

    _ee_type = 'aggregate'
    _definitions = AGGREGATES
    _exception = NoAggregate





class ElasticQueryException(Exception):
    pass


class DslException(ElasticQueryException):
    pass


class NoQuery(DslException):
    pass


class NoFilter(DslException):
    pass


class NoAggregate(DslException):
    pass


class NoSuggester(DslException):
    pass


class InvalidArg(DslException):
    pass


class MissingArg(DslException):
    pass



# coding: utf-8

from toute.exceptions import ClientError


def validate_client(es):
    """
    A valid ES client is a interface which must implements at least
    "index" and "search" public methods.
    preferably an elasticsearch.ElasticSearch() instance
    :param es:
    :return: None
    """

    if not es:
        raise ClientError("ES client cannot be Nonetype")

    try:
        if not callable(es.index) or not callable(es.search) or \
                not callable(es.get):
            raise ClientError(
                "index or search or get Interface is not callable"
            )
    except AttributeError as e:
        raise ClientError(str(e))


class FieldValidator(object):
    def __init__(self):
        self.validation = []

    def validate_value(self, field, value):
        pass

    def validate_item(self, field, item):
        pass

    def __call__(self, field, value):
        self.validate_value(field, value)
        if field._multi:
            [self.validate_item(field, item) for item in value]
        return self.validation



from toute.bases.py3 import *  # noqa
from toute.exceptions import RequiredField, InvalidMultiField
from toute.exceptions import FieldTypeMismatch, ValidationError

from collections import Iterable


class BaseField(object):
    _type = unicode
    _default = None
    _default_mapping = {'type': 'string'}

    def __init__(self, field_type=None, required=False, multi=False,
                 field_name=None, validators=None, mapping=None,
                 default=None, **kwargs):
        self._validators = validators or []
        self._field_name = field_name
        self._mapping = mapping or {}

        if field_type is not None:
            self._type = field_type

        self._required = required or getattr(self, '_required', False)
        self._multi = multi or getattr(self, '_multi', False)

        if default:
            self._default = default
        elif self._multi:
            self._default = []

        for key, value in kwargs.items():
            setattr(self, key, value)

    def validate_field_type(self, value):
        if value is not None:
            if not isinstance(value, self._type):
                raise FieldTypeMismatch(self._field_name,
                                        self._type,
                                        value.__class__)

    def validate(self, value):
        if value is None:
            if self._required:
                raise RequiredField(self._field_name)
        else:
            if self._multi:
                if not isinstance(value, Iterable):
                    raise InvalidMultiField(self._field_name)
                [self.validate_field_type(elem) for elem in value]
            else:
                self.validate_field_type(value)

        for validator in self._validators:
            """
            Functions in self._validators receives field_instance, value
            should return None or
            raise Exception (ValidationError) or return any value
            """
            val = validator(self, value)
            if val:
                raise ValidationError(
                    'Invalid %s, returned: %s' % (self._field_name, val)
                )

    def to_dict(self, value, validate=True):
        """
        Transform value from Python to be saved in E.S
        :param value: raw value
        :param validate: if should validate before transform
        :return: pure value
        """
        if validate:
            self.validate(value)
        return value

    def from_dict(self, serialized):
        """
        Transform data read from E.S to Python Object
        :param serialized: Result from E.S (string)
        :return: Instance or Instances of self._type
        """
        if serialized is not None:
            if self._multi:
                return [
                    self._type(x) if x is not None else x for x in serialized
                ]
            return self._type(serialized)
        return self._default

    @property
    def mapping(self):
        m = dict(**self._default_mapping)
        m.update(self._mapping)
        return m



from elasticsearch import Elasticsearch
from toute import Document, KeywordField, Payload, Query, Pagination


class Doc(Document):
    _index = 'test'
    _doctype = 'doc'
    _es = Elasticsearch()
    name = KeywordField()


payload = Payload(Doc, query=Query.match_all())
pagination = Pagination(payload, page=1, per_page=5)



class ClientError(Exception):
    pass


class RequiredField(Exception):
    pass


class InvalidMultiField(Exception):
    pass


class ValidationError(Exception):
    pass


class PaginationError(Exception):
    pass


class PayloadError(Exception):
    pass


class StopPagination(Exception):
    pass


class FieldTypeMismatch(Exception):

    def __init__(self, field_name, expected_type, actual_type):
        message = "`{}` expected `{}`, actual `{}`".format(
            field_name, expected_type, actual_type)
        Exception.__init__(self, message)



from toute.utils.payload.meta import BaseFilterQuery, MetaFilterQuery
from toute.utils.payload.exception import NoQuery

from six import with_metaclass

QUERIES = {
    'match': {
        'field': True,
        'args': ('query',),
        'kwargs': ('operator', 'zero_terms_query', 'cutoff_frequency', 'boost')
    },
    'multi_match': {
        'args': ({'fields': []}, 'query')
    },
    'bool': {
        'kwargs': ({('must', 'must_not', 'should'): ['_query']},)
    },
    'boost': {
        'kwargs': ({('positive', 'negative'): '_query'})
    },
    'common': {
        'args': ('query',),
        'process': lambda q: {'body': q}
    },
    'constant_score': {
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'dis_max': {
        'args': ({'queries': ['_query']},)
    },
    'filtered': {
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'fuzzy_like_this': {
        'args': ({'fields': []}, 'like_text')
    },
    'fuzzy_like_this_field': {
        'field': True,
        'args': ('like_text',),
        'kwargs': (
            'max_query_terms', 'ignore_tf', 'fuzziness', 'prefix_length',
            'boost', 'analyzer'
        )
    },
    'function_score': {
        'args': ({'functions': []},),
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'fuzzy': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost', 'fuzziness', 'prefix_length', 'max_expansions')
    },
    'geo_shape': {
        'field': True,
        'kwargs': ('type', {'coordinates': []}),
        'field_process': lambda q: {'shape': q}
    },
    'has_child': {
        'args': ('type',),
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'has_parent': {
        'args': ('parent_type',),
        'kwargs': ({'query': '_query', 'filter': '_filter'},)
    },
    'ids': {
        'args': ({'values': []},),
        'kwargs': ('type',)
    },
    'indices': {
        'args': ({'indices': []},),
        'kwargs': ({('query', 'no_match_query'): '_query'},)
    },
    'match_all': {
        'kwargs': ('boost',)
    },
    'more_like_this': {
        'args': ({'fields': []}, 'like_text')
    },
    'nested': {
        'args': ('path', {'query': '_query'}),
    },
    'prefix': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost',)
    },
    'query_string': {
        'args': ('query',),
        'kwargs': ({'fields': []},)
    },
    'simple_query_string': {
        'args': ('query',),
        'kwargs': ({'fields': []},)
    },
    'range': {
        'field': True,
        'kwargs': ('gte', 'gt', 'lte', 'lt',)
    },
    'regexp': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost', 'flags')
    },
    'span_first': {
        'args': ({'match': '_query'},)
    },
    'span_multi': {
        'args': ({'match': '_query'},)
    },
    'span_near': {
        'args': ({'clauses': ['_query']},)
    },
    'span_not': {
        'kwargs': ({('include', 'exclude'): '_query'},)
    },
    'span_or': {
        'args': ({'clauses': ['_query']},)
    },
    'span_term': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost',)
    },
    'term': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost',)
    },
    'terms': {
        'field': True,
        'value_only': True,
        'args': ({'value': ['']},)
    },
    'top_children': {
        'args': ('type',),
        'kwargs': ({'query': '_query'},)
    },
    'wildcard': {
        'field': True,
        'args': ('value',),
        'kwargs': ('boost',)
    }
}


class Query(with_metaclass(MetaFilterQuery, BaseFilterQuery)):
    _ee_type = 'query'
    _definitions = QUERIES
    _exception = NoQuery



__version__ = '0.1.0'

from toute.embedded_document import EmbeddedDocument  # noqa
from toute.document import Document  # noqa
from toute.mapping import Mapping  # noqa
from toute.fields import *  # noqa
from toute.exceptions import *  # noqa
from toute.utils.payload import Payload, Query, Filter, Aggregate, Suggester  # noqa
from toute.utils.pagination import Pagination  # noqa



# coding: utf-8
import os
import re

try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
    long_description = "ElasticSearch ODM inspired by MongoEngine, for docs please visit the homepage."


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return open(fpath(fname)).read()

# grep toute/__init__.py since python 3.x cannot import it
file_text = read(fpath('lib/toute/__init__.py'))


def grep(attrname):
    pattern = r"{0}\W*=\W*'([^']+)'".format(attrname)
    strval, = re.findall(pattern, file_text)
    return strval

setup(
    name='toute',
    version="1.1.0",
    url='https://github.com/eshta/toute',
    license='MIT',
    author="Omar Shaban",
    author_email="omars@php.net",
    description='ElasticSearch ODM inspired by MongoEngine',
    long_description=long_description,
    packages=find_packages("lib"),
    package_dir={"": "lib"},
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=["python-dateutil", "six>=1.12.0"],
    tests_require=[
        "pytest==2.8.3",
        "pytest-cov==2.2.0",
        "flake8==2.5.0",
        "pep8-naming==0.3.3",
        "flake8-debugger==1.4.0",
        "flake8-print==2.0.1",
        "flake8-todo==0.4",
        "radon==1.2.2"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ]
)



from toute.utils.payload.meta import BaseSuggester, MetaSuggester
from toute.utils.payload.exception import NoSuggester

SUGGESTERS = {
    'term': {
        'args': ('field', ),
        'kwargs': ('analyzer', 'size', 'sort', 'suggest_mode')
    },
    'phrase': {
        'args': ('field', ),
        'kwargs': (
            'gram_size', 'real_word_error_likelihood', 'confidence',
            'max_errors', 'separator', 'size', 'analyzer', 'shard_size',
            'collate'
        )
    },
    'completion': {
        'args': ('field', ),
        'kwargs': ('size', )
    }
}


class Suggester(BaseSuggester):
    __metaclass__ = MetaSuggester

    _ee_type = 'suggester'
    _definitions = SUGGESTERS
    _exception = NoSuggester



from toute.exceptions import PayloadError
from toute.utils.payload.queries import Query
from toute.utils.payload.meta_util import unroll_struct
from toute.utils.pagination import Pagination


class Payload(object):

    def __init__(self, model=None, **kwargs):
        """
        Optional parameters
        :param model: a Document model class (optional)
        :param query: A Query instance
        :param filter: A Filter instance
        :param aggregate: Aggregate instances
        :param suggest: Suggester instances
        :param sort: field name or dictionary
        :param size: Integer size
        :param timeout: Timeout in seconds
        :param fields: List of fields
        :return: Payload Wrapper
        """
        self._model = model
        self._filter = None
        self._query = None
        self._aggs = []
        self._suggesters = []
        self._struct = {}

        for key, value in kwargs.items():
            try:
                getattr(self, key)(value)
            except AttributeError:
                self.set(key, value)

    def query(self, query):
        self._query = query
        return self

    def filter(self, filter_):
        self._filter = filter_
        return self

    def aggregate(self, aggregates):
        self._aggs.extend(aggregates)
        return self

    def suggest(self, *suggesters):
        self._suggesters.extend(suggesters)
        return self

    def set(self, key, value):
        self._struct[key] = value
        return self

    def from_(self, from_):
        self._struct['from'] = from_
        return self

    def size(self, size):
        self._struct['size'] = size
        return self

    def timeout(self, timeout):
        self._struct['timeout'] = timeout
        return self

    def fields(self, fields):
        self._struct['_source'] = fields
        return self

    def sort(self, field, reset=False, **kwargs):
        """
        Sort he Payload
        :param field: Field to sort
        :param reset: Should reset sort list
        :param kwargs: "order" and other sort params
        :return: Payload instance (self)
        """
        if reset or 'sort' not in self._struct:
            self._struct['sort'] = []

        if not kwargs:
            self._struct['sort'].append(field)
        else:
            self._struct['sort'].append({field: kwargs})

        return self

    @property
    def dict(self):
        return self.as_dict()

    def as_dict(self):
        if self._filter and self._query:
            self._struct['query'] = Query.filtered(
                filter=self._filter,
                query=self._query
            )

        elif self._filter:
            self._struct['query'] = Query.filtered(
                filter=self._filter
            )

        elif self._query:
            self._struct['query'] = self._query

        if self._aggs:
            aggs = {}
            for agg in self._aggs:
                aggs.update(agg.as_dict())

            self._struct['aggregations'] = aggs

        if self._suggesters:
            suggs = {}
            for sugg in self._suggesters:
                suggs.update(sugg.as_dict())

            self._struct['suggest'] = suggs

        return unroll_struct(self._struct)

    # backward compatibility API
    to_dict = as_dict  # noqa

    def search(self, model=None, **kwargs):
        model = model or self._model
        query = self.dict
        if not query:
            raise PayloadError(
                "query, filter, aggregate or suggest should be specified!"
            )
        return model.search(query=query, **kwargs)

    def count(self, model=None, **kwargs):
        model = model or self._model
        query = self.dict.get('query')
        if not query:
            raise PayloadError("query should be specified for count")
        kwargs['perform_count'] = True
        return model.search(query={"query": query}, **kwargs)

    def paginate(self, page=1, per_page=10):
        return Pagination(iterable=self, page=page, per_page=per_page)

    def get_values(self, *fields, **kwargs):
        """
        if args is only one field .get_values('id') return a list of lists
        [123, 456, 789]
        If args is more than one field return a list of tuples
        .get_values("id", "name")
        [(123, "John"), (789, "mary"), ...]
        :param kwargs: Document class
        :param fields: a list of fields
        :return:
        """

        values = [
            hit.get('_source')
            for hit in self.search(_source=fields, **kwargs)._hits
        ]

        if not fields:
            raise AttributeError("At least one field is required")

        if len(fields) > 1:
            return [
                tuple(value.get(field) for field in fields)
                for value in values
            ]
        else:
            return [value.get(fields[0]) for value in values]



# coding: utf-8
from toute.bases.py3 import *  # noqa
from dateutil import parser
from datetime import datetime
from six import string_types
from toute.bases.field import BaseField
from toute.exceptions import ValidationError, FieldTypeMismatch
from toute.utils.validation import FieldValidator

__all__ = [
    'IntegerField', 'LongField', 'KeywordField', 'FloatField',
    'DateField', 'UuidField', 'BooleanField', 'GeoPointField',
    'ArrayField', 'ObjectField', 'TextField'
]

class TextField(BaseField):
    _type = unicode
    _default_mapping = {"index": "true", "store": "true", 'type': 'text'}

class IntegerField(BaseField):
    _type = int
    _default_mapping = {'type': 'integer'}


class LongField(BaseField):
    _type = long
    _default_mapping = {'type': 'long'}


class UuidField(BaseField):
    _type = unicode
    _default_mapping = {"store": "true", 'type': 'keyword'}


class KeywordField(BaseField):
    _type = unicode
    _default_mapping = {"index": "true", "store": "true", 'type': 'keyword'}


class FloatField(BaseField):
    _type = float
    _default_mapping = {'type': 'float'}


class BooleanField(BaseField):
    _type = bool
    _default_mapping = {'type': 'boolean'}


class ObjectField(BaseField):
    """
    Represent a typed or schema-less object (a python dict {})
    A mapping can be optionally defined in mapping argument
    example:

    >>> field = ObjectField(
    ...    mapping={"dynamic": False,
    ...             "properties": {"name": {"type": "string"}}
    ... )

    The above field will not store arbitrary properties and will accepts
    only string type in name property

    If multi=True the mapping type will be changed from 'object' to 'nested'

    If you need a more complex definition with fields and validators please
    take a look at embedded_document.EmbeddedDocument
    """
    _type = dict

    def __init__(self, *args, **kwargs):
        properties = kwargs.pop('properties', None)
        dynamic = kwargs.pop('dynamic', None)
        self._default_mapping = {'type': 'object'}
        self._default = {}
        super(ObjectField, self).__init__(*args, **kwargs)
        if dynamic is not None:
            self._default_mapping['dynamic'] = dynamic
        if properties is not None:
            self._default_mapping['properties'] = properties
        if self._multi:
            self._default_mapping['type'] = 'nested'


class ArrayField(BaseField):
    """
    ArrayField is by default a string type allowing multiple items of
    any type to be stored and retrieved as string

    It can be configured to use any of other fields as its type

    # to store an array of any objects as string
    field = ArrayField()

    # To store an array of integers (Float, Long etc)
    field = ArrayField(IntegerField())

    # As ArrayField is multi by default, if an ObjectField is used, the type
    # is turned in to 'nested' type to allow better searches.

    An array of arbitrary schema-less objects
    field = ArrayField(ObjectField())

    # equivalent to

    field = Arrayfield(field_type=dict, mapping={"type": "nested"})

    Or an array of schema strict documents

    >>> field = ArrayField(
    ...    ObjectField(
    ...        dynamic=False,
    ...        properties={"name": {"type": "string"}}
    ...    )
    ... )

    # NOTE: Schema validation is done only at E.S indexing level

    """
    _multi = True

    def __init__(self, field=None, *args, **kwargs):
        self.field = field
        self._default_mapping = {'type': 'string'}
        self._type = unicode
        if field:
            if isinstance(field, ObjectField):
                self.field._default_mapping['type'] = 'nested'
            self._default_mapping.update(self.field.mapping)
            self._type = field._type

        if 'default' not in kwargs:
            kwargs['default'] = []

        super(ArrayField, self).__init__(*args, **kwargs)

    def from_dict(self, serialized):
        """
        Transform data read from E.S to Python Object
        :param serialized: Result from E.S (string)
        :return: Instance or Instances of self._type
        """
        if serialized is not None:
            return [
                self.field.from_dict(x)
                if x is not None
                else x
                for x in serialized
            ]
        return self._default


class GeoPointStringValidator(FieldValidator):
    @staticmethod
    def validate_string(field, value):
        if value:
            values = [
                float(item.strip())
                for item in value.split(',')
            ]
            if not len(values) == 2:
                raise ValidationError(
                    '2 elements "lat,lon" required in %s' %
                    field._field_name
                )

    def validate_value(self, field, value):
        if not field._multi:
            self.validate_string(field, value)

    def validate_item(self, field, item):
        self.validate_string(field, item)


class GeoPointDictValidator(FieldValidator):
    @staticmethod
    def validate_dict(field, value):
        if value:
            for key in 'lat', 'lon':
                if not isinstance(value.get(key), float):
                    raise ValidationError(
                        '%s: %s requires a float' %
                        (field._field_name, key)
                    )

    def validate_value(self, field, value):
        if not field._multi:
            self.validate_dict(field, value)

    def validate_item(self, field, item):
        self.validate_dict(field, item)


class GeoPointField(BaseField):
    """
    A field to hold GeoPoint

    mode = dict|array|string

    >>> location = GeoPointField(mode='dict')  # default
    An object representation with lat and lon explicitly named
    >>> location = {"lat": 40.722, "lon": -73.989}}

    >>> location = GeoPointField(mode='string')
    A string representation, with "lat,lon"
    >>> location = "40.715, -74.011"

    >>> location = GeoPointField(mode='array')
    An array representation with [lon,lat].
    >>> location = [-73.983, 40.719]
    """

    def __init__(self, *args, **kwargs):
        self._default_mapping = {'type': 'geo_point'}
        self.mode = kwargs.pop('mode', 'dict')
        super(GeoPointField, self).__init__(*args, **kwargs)
        if self.mode == 'string':
            self._type = unicode
            self._validators.append(GeoPointStringValidator())

        elif self.mode == 'array':
            self._multi = True
            self._type = float
            self._default = []
            self._validators.append(_array_validator)

        else:
            self._type = dict
            self._default = {}
            self._validators.append(GeoPointDictValidator())

    def validate_field_type(self, value):
        if self.mode == 'array' and isinstance(value, list):
            def validate(val):
                if not isinstance(val, self._type):
                    raise FieldTypeMismatch(self._field_name,
                                            self._type,
                                            val.__class__)

            if value is not None:
                if any([isinstance(item, list) for item in value]):
                    [validate(item) for item in value]
        else:
            super(GeoPointField, self).validate_field_type(value)


def _validate_array_item(field, value):
    if value:
        if not len(value) == 2:
            raise ValidationError(
                '2 elements [lon, lat] required in %s' %
                field._field_name
            )


def _array_validator(field, value):
    if any([isinstance(item, list) for item in value]):
        # it is a multi location geo array
        [_validate_array_item(field, item) for item in value]
    else:
        _validate_array_item(field, value)


class DateField(BaseField):
    _type = datetime
    _default_mapping = {"type": "date"}

    @property
    def _date_format(self):
        """
        Optional string format used to send date value to E.S
        specified in DateField(date_format="%Y-%m-%d %H:%M:%S")
        if not specified isoformat() will be used
        :return: string date format or None
        """
        return getattr(self, 'date_format', None)

    def to_dict(self, value, validate=True):
        if self._multi:
            if not value:
                return []
            self.validate(value)
            if self._date_format:
                return [x.strftime(self._date_format) for x in value]
            return [x.isoformat() for x in value]
        else:
            if not value:
                return None
            if validate:
                self.validate(value)
            if self._date_format:
                return value.strftime(self._date_format)
            return value.isoformat()

    def from_dict(self, serialized):
        if serialized:
            if self._multi:
                values = []
                for elem in serialized:
                    if elem is None:
                        continue
                    if isinstance(elem, self._type):
                        values.append(elem)
                    elif isinstance(elem, string_types):
                        date = parser.parse(elem)
                        values.append(date)
                    else:
                        raise ValueError(
                            'Expected str or date. {} found'.format(
                                elem.__class__
                            )
                        )
                return values
            else:
                if serialized is None:
                    return None
                if isinstance(serialized, self._type):
                    return serialized
                elif isinstance(serialized, string_types):
                    return parser.parse(serialized)
                raise ValueError('Expected str or date. {} found'.format(
                    serialized.__class__)
                )



from toute.bases.py3 import *  # noqa
from toute.utils.payload.exception import InvalidArg, MissingArg
from six import iteritems


def _check_input(arg):
    if arg is None:
        return False

    if type(arg) in (list, dict, tuple) and len(arg) == 0:
        return False

    return True


def _check_type(key, type_, arg):
    if isinstance(type_, basestring):
        # Empty string means string type
        if not type_:
            type_ = str
        elif not hasattr(arg, '_ee_type') or arg._ee_type != type_[1:]:
            raise InvalidArg(
                '{0} should be a {1}'.format(key, type_[1:].title())
            )
    elif not isinstance(arg, type_):
        raise InvalidArg('{0} should be a list of {1}'.format(key, type_))


def _check_arg(key, expected_type, arg):
    if isinstance(expected_type, list):
        if not isinstance(arg, list):
            raise InvalidArg('{0} should be a list'.format(key))

        if expected_type:
            # Loop the list and check all it's args
            for arg in arg:
                _check_type(key, expected_type[0], arg)

    elif expected_type is not None:
        _check_type(key, expected_type, arg)

    else:
        if isinstance(arg, dict) or isinstance(arg, list):
            raise InvalidArg('{0} should be a string or integer'.format(key))


def _parse_args(args, argspec):
    struct = {}
    arg_length = len(args)

    for i, (key, expected_type) in enumerate(argspec):
        if i >= arg_length:
            raise MissingArg('Missing {0} argument'.format(key))

        if _check_input(args[i]):
            _check_arg(key, expected_type, args[i])
            struct[key] = args[i]

    return struct


def _parse_kwargs(kwargs, kwargspec):
    struct = {}

    for (key, expected_type) in kwargspec:
        if key in kwargs and _check_input(kwargs[key]):
            _check_arg(key, expected_type, kwargs[key])
            struct[key] = kwargs.pop(key)

    return struct


def make_struct(definition, *args, **kwargs):
    if isinstance(definition, list):
        _check_arg('', definition, list(args))
        struct = args
    else:
        struct = {}
        # Field type ({field: {args}})
        if definition.get('field'):
            field_args = _parse_args(args[1:], definition.get('args', []))
            field_kwargs = _parse_kwargs(kwargs, definition.get('kwargs', []))

            # Value must be set in definition args, for queries
            # where no kwargs are allowed
            if definition.get('value_only'):
                field_struct = field_args['value']

            # No kwargs and value set?
            elif not field_kwargs and 'value' in field_args:
                field_struct = field_args['value']
            else:
                field_struct = {}
                field_struct.update(field_args)
                field_struct.update(field_kwargs)

            if 'field_process' in definition:
                field_struct = definition['field_process'](field_struct)

            struct[args[0]] = field_struct

        # Normal type ({args})
        else:
            struct.update(_parse_args(args, definition.get('args', [])))
            struct.update(_parse_kwargs(kwargs, definition.get('kwargs', [])))

        # Update any remaining kwargs (excluding Nones)
        struct.update({
            k: v for k, v in iteritems(kwargs)
            if _check_input(v)
        })

    # Post-process the struct
    if 'process' in definition:
        struct = definition['process'](struct)

    return struct


def unroll_struct(struct):
    if type(struct) in (list, tuple):
        return [unroll_struct(v) for v in struct]
    elif isinstance(struct, dict):
        return {k: unroll_struct(v) for k, v in iteritems(struct)}
    elif getattr(struct, '_ee_type', None):
        return unroll_struct(struct.as_dict())
    else:
        return struct


def _unroll_spec(spec):
    out_spec = []

    for arg in spec:
        if isinstance(arg, dict):
            for key, expected_type in iteritems(arg):
                if isinstance(key, tuple):
                    for sub_key in key:
                        out_spec.append((sub_key, expected_type))
                else:
                    out_spec.append((key, expected_type))
        else:
            out_spec.append((arg, None))

    return out_spec


def unroll_definitions(definitions):
    for key, spec in iteritems(definitions):
        if isinstance(spec, dict):
            for arg_type in ['args', 'kwargs']:
                if arg_type in spec:
                    spec[arg_type] = _unroll_spec(spec[arg_type])
